package com.hidglobal.omnikeydemo

import android.os.AsyncTask
import android.smartcardio.Card
import android.smartcardio.CardException
import android.smartcardio.CardTerminal
import com.hidglobal.omnikeydemo.Constants.WILDCARD_PROTOCOL
import com.hidglobal.omnikeydemo.extensions.sendAPDU
import com.hidglobal.omnikeydemo.model.CardCommunicationProgress
import com.hidglobal.omnikeydemo.model.CardCommunicationStep
import timber.log.Timber
import java.io.UnsupportedEncodingException
import java.lang.ref.WeakReference
import java.nio.charset.Charset

internal class GetCardStatusTask(demoActivity: OmniKeyDemoActivity) : AsyncTask<CardTerminal, CardCommunicationProgress, Void>() {

    private val demoActivityRef: WeakReference<OmniKeyDemoActivity> = WeakReference(demoActivity)

    public override fun doInBackground(vararg params: CardTerminal): Void? {
        val terminals = params.toList()
        while (!isCancelled) {
            try {
                publishProgress(CardCommunicationProgress(CardCommunicationStep.CLEAR_UI))
                publishProgress(CardCommunicationProgress(CardCommunicationStep.CARD_STATUS, demoActivityRef.get()?.getString(R.string.waiting_for_card).orEmpty()))

                terminals.forEach {
                    Timber.d("Waiting on terminal %s", it.name)
                    if (it.isCardPresent) {
                        Timber.d("Card present on terminal %s", it.name)
                        communicateWithCardAndUpdateUi(it)
                        Timber.d("Card can be removed from terminal %s", it.name)
                    }
                }
                Thread.sleep(USB_SLEEP_MILLIS)
            } catch (e: Exception) {
                Timber.e(e)
            }

        }
        return null
    }

    public override fun onProgressUpdate(vararg params: CardCommunicationProgress) {
        if (params.isNotEmpty()) {
            val cp = params[0]
            when (cp.step) {
                CardCommunicationStep.CLEAR_UI -> demoActivityRef.get()?.clearUI()
                CardCommunicationStep.CARD_STATUS -> demoActivityRef.get()?.setCardStatusText(cp.message)
                CardCommunicationStep.ATR_CMD -> demoActivityRef.get()?.setAtrText(cp.message)
                CardCommunicationStep.UID_CMD -> demoActivityRef.get()?.setUidText(cp.message)
                CardCommunicationStep.FW_CMD -> demoActivityRef.get()?.setFirmwareText(cp.message)
                CardCommunicationStep.READER_NAME_CMD -> demoActivityRef.get()?.setReaderNameText(cp.message)
            }
        }
    }

    public override fun onCancelled(unused: Void?) {
        demoActivityRef.get()?.clearUI()
    }

    @Throws(CardException::class)
    private fun communicateWithCardAndUpdateUi(reader: CardTerminal) {
        publishProgress(CardCommunicationProgress(CardCommunicationStep.CARD_STATUS, demoActivityRef.get()?.getString(R.string.card_present).orEmpty()))
        val card = reader.connect(WILDCARD_PROTOCOL)
        publishProgress(CardCommunicationProgress(CardCommunicationStep.CARD_STATUS, demoActivityRef.get()?.getString(R.string.getting_data).orEmpty()))
        publishProgress(CardCommunicationProgress(CardCommunicationStep.ATR_CMD, getCardAtr(card)))
        publishProgress(CardCommunicationProgress(CardCommunicationStep.READER_NAME_CMD, getReaderName(card, reader)))
        publishProgress(CardCommunicationProgress(CardCommunicationStep.UID_CMD, getUid(card)))
        publishProgress(CardCommunicationProgress(CardCommunicationStep.FW_CMD, getFirmwareInfo(card)))
        publishProgress(CardCommunicationProgress(CardCommunicationStep.CARD_STATUS))
        card.disconnect(true)
        var isCardAbsent = false
        while (!isCancelled && !isCardAbsent) {
            isCardAbsent = reader.waitForCardAbsent(CARD_ABSENT_MILLIS)
            Thread.sleep(THREAD_SLEEP_MILLIS)
        }
    }

    private fun getFirmwareInfo(card: Card): String {
        var fwInfo: String
        val response = card.sendAPDU(GET_FIRMWARE_INFO_COMMAND)
        if (response != null) {
            val data = ConvertionUtils.extractParameters(response)
            if (data != null && !data.isEmpty()) {
                return try {
                    fwInfo = String(data[0], Charset.forName(CHARSET_ASCII))
                    fwInfo = fwInfo.substring(0, fwInfo.lastIndexOf('-'))
                    fwInfo = fwInfo.substring(0, fwInfo.lastIndexOf('-'))
                    fwInfo
                } catch (e: UnsupportedEncodingException) {
                    cancel(true)
                    ""
                }
            }
        }
        return ""
    }

    private fun getUid(card: Card): String {
        var uid: String? = null
        val response = card.sendAPDU(GET_UID_COMMAND)
        response?.let { uid = ConvertionUtils.byteArrayToString(it, 2) }
        return uid.orEmpty()
    }

    private fun getReaderName(card: Card, cardTerminal: CardTerminal): String {
        val response = card.sendAPDU(GET_READER_NAME_COMMAND)
        if (response != null) {
            val data = ConvertionUtils.extractParameters(response)
            if (data != null && !data.isEmpty()) {
                return try {
                    String(data[0], Charset.forName(CHARSET_ASCII))
                } catch (e: UnsupportedEncodingException) {
                    cancel(true)
                    cardTerminal.name
                }

            }
        }
        return cardTerminal.name
    }

    private fun getCardAtr(card: Card): String {
        var atr: String? = null
        val respATR = card.atr
        respATR?.let { atr = ConvertionUtils.byteArrayToString(it.bytes, 0) }
        return atr.orEmpty()
    }

    companion object {
        private const val GET_FIRMWARE_INFO_COMMAND = "FF70076B08A206A004A002960000"
        private const val GET_UID_COMMAND = "FFCA000000"
        private const val GET_READER_NAME_COMMAND = "FF70076B08A206A004A002820000"
        private const val CHARSET_ASCII = "ASCII"
        private const val USB_SLEEP_MILLIS = 200L
        private const val CARD_ABSENT_MILLIS = 100L
        private const val THREAD_SLEEP_MILLIS = 100L

    }
}
