package com.hidglobal.omnikeydemo

import android.app.AlertDialog
import android.content.ComponentName
import android.content.IntentFilter
import android.content.ServiceConnection
import android.content.pm.PackageManager
import android.os.AsyncTask
import android.os.Build
import android.os.Bundle
import android.os.IBinder
import android.smartcardio.CardException
import android.smartcardio.CardTerminal
import android.smartcardio.TerminalFactory
import android.smartcardio.hidglobal.Constants
import android.smartcardio.hidglobal.PackageManagerQuery
import android.smartcardio.ipc.CardService
import android.smartcardio.ipc.ICardService
import android.support.v4.app.ActivityCompat
import android.support.v4.content.ContextCompat
import android.support.v7.app.AppCompatActivity
import android.view.View
import android.widget.Toast
import com.hidglobal.omnikeydemo.extensions.hideKeyboard
import com.hidglobal.omnikeydemo.model.CardCommunicationProgress
import com.hidglobal.omnikeydemo.usb.UsbEventsListener
import com.hidglobal.omnikeydemo.usb.UsbEventsReceiver
import kotlinx.android.synthetic.main.activity_omnikey_demo.*
import timber.log.Timber

class OmniKeyDemoActivity : AppCompatActivity() {

    private var cardService: ICardService? = null
    private var terminalFactory: TerminalFactory? = null
    private var availableCardTerminals = listOf<CardTerminal>()
    private var chosenTerminalIndex = 0
    private var pollTerminalsTask: AsyncTask<CardTerminal, CardCommunicationProgress, Void>? = null
    private var cardResponseTask: AsyncTask<CardTerminal, CardCommunicationProgress, Void>? = null
    private var usbEventsReceiver: UsbEventsReceiver? = null
    private var isServiceConnected = false
    private var isPermissionAskingInProgress = false

    public override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        setContentView(R.layout.activity_omnikey_demo)
        clearUI()

        generateApduResponseBtn.isEnabled = false
        generateApduResponseBtn.setOnClickListener {
            availableCardTerminals = getAvailableCardTerminals()
            if (availableCardTerminals.isNotEmpty()) {
                if (availableCardTerminals.size > 1) {
                    chooseTerminal(availableCardTerminals)
                } else {
                    cardResponseTask = GetApduResponseTask(this@OmniKeyDemoActivity).execute(availableCardTerminals[0])
                }
            }
            it.hideKeyboard()
        }
    }

    override fun onResume() {
        super.onResume()
        // if card reader manager is installed get card service from CardReaderManager app
        if (PackageManagerQuery().isCardManagerAppInstalled(this)) {
            if (ContextCompat.checkSelfPermission(this, Constants.PERMISSION_TO_BIND_BACKEND_SERVICE) == PackageManager.PERMISSION_GRANTED) {
                bindCardService()
            } else {
                if (!isPermissionAskingInProgress) {
                    requestruntimePermissions()
                }
            }
        } else {
            showReaderAppNotInstalledDialog()
        }
    }

    override fun onPause() {
        super.onPause()
        cardService?.releaseService()
    }

    private fun chooseTerminal(availableTerminals: List<CardTerminal>) {
        return this.let {
            AlertDialog.Builder(it)
                    .setTitle(R.string.choose_a_reader)
                    .setCancelable(false)
                    .setSingleChoiceItems(availableTerminals.map { it.name }.toTypedArray(), noneOfItemsSelectedByDefault) { _, which ->
                        chosenTerminalIndex = which
                    }
                    .setPositiveButton(R.string.ok) { _, _ ->
                        cardResponseTask = GetApduResponseTask(this).execute(availableCardTerminals[chosenTerminalIndex])
                    }
                    .setNegativeButton(R.string.cancel) { _, _ ->
                        chosenTerminalIndex = 0
                    }
                    .create()
                    .show()
        }
    }

    private fun requestruntimePermissions() {
        ActivityCompat.requestPermissions(this, arrayOf(Constants.PERMISSION_TO_BIND_BACKEND_SERVICE), REQUEST_BIND_BACKEND_SERVICE_PERMISSION)
        isPermissionAskingInProgress = true
    }

    override fun onStart() {
        super.onStart()
        registerUsbEventsReceiver()
    }

    public override fun onStop() {
        super.onStop()
        pollTerminalsTask?.cancel(true)
        cardResponseTask?.cancel(true)
        usbEventsReceiver?.listener = null
        try {
            unregisterReceiver(usbEventsReceiver)
        } catch (e: Exception) {
            Timber.e(e)
        }
    }

    private fun showReaderAppNotInstalledDialog() {
        AlertDialog.Builder(this)
                .setTitle(getString(R.string.install_card_reader_manager_app))
                .setMessage(getString(R.string.install_card_reader_manager))
                .setPositiveButton(R.string.ok) { _, _ -> finish() }
                .setCancelable(false)
                .show()
    }

    private fun registerUsbEventsReceiver() {
        val broadcastFilter = IntentFilter()
        broadcastFilter.addAction(UsbEventsReceiver.READER_ATTACH_ACTION)
        broadcastFilter.addAction(UsbEventsReceiver.READER_DETACH_ACTION)
        usbEventsReceiver = UsbEventsReceiver(object : UsbEventsListener {
            override fun onUsbReaderAttached() {
                availableCardTerminals = getAvailableCardTerminals()
                getCardStatus()
                Timber.d("Usb reader attached")
            }

            override fun onUsbReaderDetach() {
                pollTerminalsTask?.cancel(true)
                cardResponseTask?.cancel(true)
                availableCardTerminals = emptyList()
                Timber.d("Usb reader detached")
            }
        })
        registerReceiver(usbEventsReceiver, broadcastFilter)
    }

    private fun bindCardService() {

        if (!isServiceConnected) {
            cardService = CardService.getInstance(this, object : ServiceConnection {
                override fun onServiceConnected(componentName: ComponentName, iBinder: IBinder) {
                    isServiceConnected = true
                    Timber.d("OmniKeyDemo connected to card service")
                    Timber.d("Reader connected")

                    generateApduResponseBtn.isEnabled = true

                    manualApduCb.setOnCheckedChangeListener { _, isChecked ->
                        if (isChecked) {
                            pollTerminalsTask?.cancel(true)
                            chooseLayoutToShow(ShowLayout.APDU)
                        } else {
                            chooseLayoutToShow(ShowLayout.MAIN)
                            pollTerminalsTask = GetCardStatusTask(this@OmniKeyDemoActivity).execute(*(availableCardTerminals.toTypedArray()))
                        }
                    }

                    if (manualApduCb.isChecked) {
                        chooseLayoutToShow(ShowLayout.APDU)
                    } else {
                        chooseLayoutToShow(ShowLayout.MAIN)
                        availableCardTerminals = getAvailableCardTerminals()
                        getCardStatus()
                    }
                }

                override fun onServiceDisconnected(componentName: ComponentName) {
                    isServiceConnected = false
                    Timber.d("OmniKeyDemo disconnected to card service")
                }
            })
        } else {
            generateApduResponseBtn.isEnabled = true
            if (!manualApduCb.isChecked) {
                availableCardTerminals = getAvailableCardTerminals()
                getCardStatus()
            }
        }
    }

    override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<String>, grantResults: IntArray) {
        if (requestCode == REQUEST_BIND_BACKEND_SERVICE_PERMISSION && grantResults.isNotEmpty()) {
            if (grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                bindCardService()
            } else {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                    AlertDialog.Builder(this)
                            .setTitle(R.string.permission_required)
                            .setMessage(R.string.app_cannot_be_run_without_permission)
                            .setPositiveButton(R.string.ok) { _, _ ->
                                isPermissionAskingInProgress = false
                                requestruntimePermissions()
                            }
                            .show()
                } else {
                    // on Android < 6 runtime permissions are not available, all permissions are granted automatically by system
                    // but PERMISSION_TO_BIND_BACKEND_SERVICE is a custom permission unknown by OS if CardReaderManager app is not installed
                    // when permission has PERMISSION_DENIED status it means that user has installed Client app first and ReaderManager app later
                    // therefore system didn't grant permission automatically, Client app must be reinstalled after ReaderManager app is installed
                    AlertDialog.Builder(this)
                            .setTitle(R.string.reinstall_required)
                            .setMessage(R.string.reinstall_required_explanation)
                            .setPositiveButton(R.string.ok) { _, _ -> finish() }
                            .show()
                }
            }
        } else {
            super.onRequestPermissionsResult(requestCode, permissions, grantResults)
        }
    }

    private fun getAvailableCardTerminals(): List<CardTerminal> {
        if (terminalFactory == null) {
            try {
                terminalFactory = cardService?.terminalFactory
            } catch (e: Exception) {
                Timber.e(e, "Unable to get terminal factory")
                return emptyList()
            }
        }

        try {
            terminalFactory?.terminals()?.list()?.let {
                return it.filterNotNull()
            }
        } catch (e: CardException) {
            Timber.e(e)
        }
        return emptyList()
    }


    fun getCardStatus() {
        if (cardService == null) {
            Toast.makeText(this, R.string.cannot_conntect_to_backend, Toast.LENGTH_LONG).show()
            Timber.e("Cannot connect to Backend Service, make sure CardReaderManager app is installed")
            return
        }
        cardResponseTask?.cancel(true)
        pollTerminalsTask?.cancel(true)
        if (availableCardTerminals.isNotEmpty()) {
            pollTerminalsTask = GetCardStatusTask(this).execute(*(availableCardTerminals.toTypedArray()))
        }
    }

    fun clearUI() = listOf(atrTv, uidTv, firmwareTv, readerNameTv, cardStatusTv, apduResponseTv).forEach { it.text = "" }

    private fun chooseLayoutToShow(layoutType: ShowLayout) {
        listOf(apduCommandEt, generateApduResponseBtn, apduResponseLbl, apduResponseTv).forEach {
            when (layoutType) {
                ShowLayout.APDU -> it.visibility = View.VISIBLE
                ShowLayout.MAIN -> it.visibility = View.GONE
            }
        }
        listOf(atrEt, atrTv, uidEt, uidTv).forEach {
            when (layoutType) {
                ShowLayout.APDU -> it.visibility = View.GONE
                ShowLayout.MAIN -> it.visibility = View.VISIBLE
            }
        }
    }

    fun setAtrText(string: String) {
        atrTv.text = string
    }

    fun setUidText(string: String) {
        uidTv.text = string
    }

    fun setFirmwareText(string: String) {
        firmwareTv.text = string
    }

    fun setReaderNameText(string: String) {
        readerNameTv.text = string
    }

    fun setCardStatusText(string: String) {
        cardStatusTv.text = string
    }

    fun setCardResponseText(string: String) {
        apduResponseTv.text = string
    }

    fun getApduRequestText() = apduCommandEt.text.toString()

    companion object {
        private const val REQUEST_BIND_BACKEND_SERVICE_PERMISSION = 9000
        private const val noneOfItemsSelectedByDefault = -1
    }
}
